@@ -42,11 +42,11 @@
  * terms and conditions of the copyright.
  */
 
-#include "h/common.h"
-#include "h/ip.h"
-#include "h/tcp.h"
-#include "h/udp.h"
-#include "h/main.h"
+#include <slirp.h>
+
+#ifndef HAVE_MEMMOVE
+#define memmove(x, y, z) bcopy(y, x, z)
+#endif
 
 int ip_defttl;
 struct ipstat ipstat;
@@ -59,7 +59,7 @@ struct ipq ipq;
 void
 ip_init()
 {
-	ipq.next = ipq.prev = &ipq;
+	ipq.next = ipq.prev = (ipqp_32)&ipq;
 	ip_id = tt.tv_sec & 0xffff;
 	udp_init();
 	tcp_init();
@@ -78,22 +78,18 @@ ip_input(m)
 	register struct ipq *fp;
 	int hlen;
 	
-#ifdef DEBUG
-	debug_call(dfd,"ip_input(m) called ...\n");
-	debug_call(dfd,"    m = %d\n", (int)m);
-	debug_call(dfd,"        m->m_len = %d\n", m->m_len);
-	fflush_call(dfd);
-#endif
-
+	DEBUG_CALL("ip_input");
+	DEBUG_ARG("m = %lx", (long)m);
+	
 	ipstat.ips_total++;
 	
 	if (m->m_len < sizeof (struct ip)) {
 		ipstat.ips_toosmall++;
 		return;
 	}
-
+	
 	ip = mtod(m, struct ip *);
-
+	
 	if (ip->ip_v != IPVERSION) {
 		ipstat.ips_badvers++;
 		goto bad;
@@ -165,7 +161,8 @@ ip_input(m)
 		 * Look for queue of fragments
 		 * of this datagram.
 		 */
-		for (fp = ipq.next; fp != &ipq; fp = fp->next)
+		for (fp = (struct ipq *) ipq.next; fp != &ipq;
+		    fp = (struct ipq *) fp->next)
 			if (ip->ip_id == fp->ipq_id &&
 			    ip->ip_src.s_addr == fp->ipq_src.s_addr &&
 			    ip->ip_dst.s_addr == fp->ipq_dst.s_addr &&
@@ -216,17 +213,10 @@ found:
 		udp_input(m, hlen);
 		break;
 	 case IPPROTO_ICMP:
-#ifdef WORDS_BIGENDIAN
-#define ECHOREQ 0x0800
-#else
-#define ECHOREQ 0x0008
-#endif
-		m->m_data += sizeof(struct ip);
-		if (*mtod(m, u_short *) == ECHOREQ)
-		   slirp_errors |= ERR_RCVICMPECHOREQ;
-		/* FALLTHROUGH */
-#undef ECHOREQ
+		icmp_input(m, hlen);
+		break;
 	 default:
+		ipstat.ips_noproto++;
 		m_free(m);
 	}
 	return;
@@ -251,12 +241,10 @@ ip_reass(ip, fp)
 	struct mbuf *t;
 	int hlen = ip->ip_hl << 2;
 	int i, next, delta;
-
-#ifdef DEBUG
-	debug_call(dfd,"ip_reass(ip,fp) called ...\n");
-	debug_call(dfd,"    ip = %d\n    fp = %d\n", (int)ip, (int)fp);
-	fflush_call(dfd);
-#endif
+	
+	DEBUG_CALL("ip_reass");
+	DEBUG_ARG("ip = %lx", (long)ip);
+	DEBUG_ARG("fp = %lx", (long)fp);
 
 	/*
 	 * Presence of header sizes in mbufs
@@ -272,11 +260,11 @@ ip_reass(ip, fp)
 		if ((t = m_get()) == NULL)
 			goto dropfrag;
 		fp = mtod(t, struct ipq *);
-		insque(fp, &ipq);
+		insque_32(fp, &ipq);
 		fp->ipq_ttl = IPFRAGTTL;
 		fp->ipq_p = ip->ip_p;
 		fp->ipq_id = ip->ip_id;
-		fp->ipq_next = fp->ipq_prev = (struct ipasfrag *)fp;
+		fp->ipq_next = fp->ipq_prev = (ipasfragp_32)fp;
 		fp->ipq_src = ((struct ip *)ip)->ip_src;
 		fp->ipq_dst = ((struct ip *)ip)->ip_dst;
 		q = (struct ipasfrag *)fp;
@@ -286,7 +274,8 @@ ip_reass(ip, fp)
 	/*
 	 * Find a segment which begins after this one does.
 	 */
-	for (q = fp->ipq_next; q != (struct ipasfrag *)fp; q = q->ipf_next)
+	for (q = (struct ipasfrag *)fp->ipq_next; q != (struct ipasfrag *)fp;
+	    q = (struct ipasfrag *)q->ipf_next)
 		if (q->ip_off > ip->ip_off)
 			break;
 
@@ -295,8 +284,9 @@ ip_reass(ip, fp)
 	 * our data already.  If so, drop the data from the incoming
 	 * segment.  If it provides all of our data, drop us.
 	 */
-	if (q->ipf_prev != (struct ipasfrag *)fp) {
-		i = q->ipf_prev->ip_off + q->ipf_prev->ip_len - ip->ip_off;
+	if (q->ipf_prev != (ipasfragp_32)fp) {
+		i = ((struct ipasfrag *)(q->ipf_prev))->ip_off +
+		  ((struct ipasfrag *)(q->ipf_prev))->ip_len - ip->ip_off;
 		if (i > 0) {
 			if (i >= ip->ip_len)
 				goto dropfrag;
@@ -318,9 +308,9 @@ ip_reass(ip, fp)
 			m_adj(dtom(q), i);
 			break;
 		}
-		q = q->ipf_next;
-		m_freem(dtom(q->ipf_prev));
-		ip_deq(q->ipf_prev);
+		q = (struct ipasfrag *) q->ipf_next;
+		m_freem(dtom((struct ipasfrag *) q->ipf_prev));
+		ip_deq((struct ipasfrag *) q->ipf_prev);
 	}
 
 insert:
@@ -328,26 +318,27 @@ insert:
 	 * Stick new segment in its place;
 	 * check for complete reassembly.
 	 */
-	ip_enq(ip, q->ipf_prev);
+	ip_enq(ip, (struct ipasfrag *) q->ipf_prev);
 	next = 0;
-	for (q = fp->ipq_next; q != (struct ipasfrag *)fp; q = q->ipf_next) {
+	for (q = (struct ipasfrag *) fp->ipq_next; q != (struct ipasfrag *)fp;
+	    q = (struct ipasfrag *) q->ipf_next) {
 		if (q->ip_off != next)
 			return (0);
 		next += q->ip_len;
 	}
-	if (q->ipf_prev->ipf_mff & 1)
+	if (((struct ipasfrag *)(q->ipf_prev))->ipf_mff & 1)
 		return (0);
 
 	/*
 	 * Reassembly is complete; concatenate fragments.
 	 */
-	q = fp->ipq_next;
+	q = (struct ipasfrag *) fp->ipq_next;
 	m = dtom(q);
 
-	q = q->ipf_next;
+	q = (struct ipasfrag *) q->ipf_next;
 	while (q != (struct ipasfrag *)fp) {
 		t = dtom(q);
-		q = q->ipf_next;
+		q = (struct ipasfrag *) q->ipf_next;
 		m_cat(m, t);
 	}
 
@@ -357,7 +348,7 @@ insert:
 	 * dequeue and discard fragment reassembly header.
 	 * Make header visible.
 	 */
-	ip = fp->ipq_next;
+	ip = (struct ipasfrag *) fp->ipq_next;
 
 	/*
 	 * If the fragments concatenated to an mbuf that's
@@ -375,7 +366,7 @@ insert:
 	ip->ipf_mff &= ~1;
 	((struct ip *)ip)->ip_src = fp->ipq_src;
 	((struct ip *)ip)->ip_dst = fp->ipq_dst;
-	remque(fp);
+	remque_32(fp);
 	(void) m_free(dtom(fp));
 	m = dtom(ip);
 	m->m_len += (ip->ip_hl << 2);
@@ -399,12 +390,13 @@ ip_freef(fp)
 {
 	register struct ipasfrag *q, *p;
 
-	for (q = fp->ipq_next; q != (struct ipasfrag *)fp; q = p) {
-		p = q->ipf_next;
+	for (q = (struct ipasfrag *) fp->ipq_next; q != (struct ipasfrag *)fp;
+	    q = p) {
+		p = (struct ipasfrag *) q->ipf_next;
 		ip_deq(q);
 		m_freem(dtom(q));
 	}
-	remque(fp);
+	remque_32(fp);
 	(void) m_free(dtom(fp));
 }
 
@@ -416,17 +408,10 @@ void
 ip_enq(p, prev)
 	register struct ipasfrag *p, *prev;
 {
-
-#ifdef DEBUG
-	debug_call(dfd,"ip_enq(p, prev) called ...\n");
-	debug_call(dfd,"    p = %d\n    prev = %d\n", (int)p, (int)prev);
-	fflush_call(dfd);
-#endif
-
-	p->ipf_prev = prev;
+	p->ipf_prev = (ipasfragp_32) prev;
 	p->ipf_next = prev->ipf_next;
-	prev->ipf_next->ipf_prev = p;
-	prev->ipf_next = p;
+	((struct ipasfrag *)(prev->ipf_next))->ipf_prev = (ipasfragp_32) p;
+	prev->ipf_next = (ipasfragp_32) p;
 }
 
 /*
@@ -436,14 +421,8 @@ void
 ip_deq(p)
 	register struct ipasfrag *p;
 {
-#ifdef DEBUG
-	debug_call(dfd,"ip_deq(p) called ...\n");
-	debug_call(dfd,"    p = %d\n", (int)p);
-	fflush_call(dfd);
-#endif
-	
-	p->ipf_prev->ipf_next = p->ipf_next;
-	p->ipf_next->ipf_prev = p->ipf_prev;
+	((struct ipasfrag *)(p->ipf_prev))->ipf_next = p->ipf_next;
+	((struct ipasfrag *)(p->ipf_next))->ipf_prev = p->ipf_prev;
 }
 
 /*
@@ -455,22 +434,19 @@ void
 ip_slowtimo()
 {
 	register struct ipq *fp;
-
-#ifdef DEBUG
-	debug_verbose(dfd,"ip_slowtimo() called ...\n");
-	fflush_call(dfd);
-#endif
-
-	fp = ipq.next;
+	
+	DEBUG_CALL("ip_slowtimo");
+	
+	fp = (struct ipq *) ipq.next;
 	if (fp == 0)
 	   return;
 
 	while (fp != &ipq) {
 		--fp->ipq_ttl;
-		fp = fp->next;
-		if (fp->prev->ipq_ttl == 0) {
+		fp = (struct ipq *) fp->next;
+		if (((struct ipq *)(fp->prev))->ipq_ttl == 0) {
 			ipstat.ips_fragtimeout++;
-			ip_freef(fp->prev);
+			ip_freef((struct ipq *) fp->prev);
 		}
 	}
 }
@@ -544,23 +520,23 @@ typedef u_int32_t n_time;
 					code = ICMP_UNREACH_SRCFAIL;
 					goto bad;
 				}
-				 *
+				/*
 				 * Loose routing, and not at next destination
 				 * yet; nothing to do except forward.
-				 *
+				 */
 				break;
 			}
 			off--;			/ * 0 origin *  /
 			if (off > optlen - sizeof(struct in_addr)) {
-				 *
+				/*
 				 * End of source route.  Should be for us.
-				 *
+				 */
 				save_rte(cp, ip->ip_src);
 				break;
 			}
-			 *
+			/*
 			 * locate outgoing interface
-			 *
+			 */
 			bcopy((caddr_t)(cp + off), (caddr_t)&ipaddr.sin_addr,
 			    sizeof(ipaddr.sin_addr));
 			if (opt == IPOPT_SSRR) {
@@ -579,9 +555,9 @@ typedef u_int32_t n_time;
 			bcopy((caddr_t)&(IA_SIN(ia)->sin_addr),
 			    (caddr_t)(cp + off), sizeof(struct in_addr));
 			cp[IPOPT_OFFSET] += sizeof(struct in_addr);
-			 *
-			 * Let ip_intr''s mcast routing check handle mcast pkts
-			 *
+			/*
+			 * Let ip_intr's mcast routing check handle mcast pkts
+			 */
 			forward = !IN_MULTICAST(ntohl(ip->ip_dst.s_addr));
 			break;
 
@@ -590,18 +566,18 @@ typedef u_int32_t n_time;
 				code = &cp[IPOPT_OFFSET] - (u_char *)ip;
 				goto bad;
 			}
-			 *
+			/*
 			 * If no space remains, ignore.
-			 *
+			 */
 			off--;			 * 0 origin *
 			if (off > optlen - sizeof(struct in_addr))
 				break;
 			bcopy((caddr_t)(&ip->ip_dst), (caddr_t)&ipaddr.sin_addr,
 			    sizeof(ipaddr.sin_addr));
-			 *
-			 * locate outgoing interface; if we''re the destination,
+			/*
+			 * locate outgoing interface; if we're the destination,
 			 * use the incoming interface (should be same).
-			 *
+			 */
 			if ((ia = (INA)ifa_ifwithaddr((SA)&ipaddr)) == 0 &&
 			    (ia = ip_rtaddr(ipaddr.sin_addr)) == 0) {
 				type = ICMP_UNREACH;
@@ -634,7 +610,7 @@ typedef u_int32_t n_time;
 				    sizeof(struct in_addr) > ipt->ipt_len)
 					goto bad;
 				ipaddr.sin_addr = dst;
-				ia = (INA)ifaof_ifpforaddr((SA)&ipaddr,
+				ia = (INA)ifaof_ i f p foraddr((SA)&ipaddr,
 							    m->m_pkthdr.rcvif);
 				if (ia == 0)
 					continue;
@@ -698,12 +674,6 @@ ip_stripoptions(m, mopt)
 	register caddr_t opts;
 	int olen;
 
-#ifdef DEBUG
-	debug_call(dfd,"ip_stripoptions(m, mopt) called ...\n");
-	debug_call(dfd,"    m = %d\n    mopt = %d\n", (int)m, (int)mopt);
-	fflush_call(dfd);
-#endif
-	
 	olen = (ip->ip_hl<<2) - sizeof (struct ip);
 	opts = (caddr_t)(ip + 1);
 	i = m->m_len - (sizeof (struct ip) + olen);
