@@ -42,12 +42,7 @@
  * terms and conditions of the copyright.
  */
 
-#include "h/common.h"
-#include "h/ip.h"
-#include "h/tcp.h"
-#include "h/socket.h"
-#include "h/main.h"
-#include "h/ctl.h"
+#include <slirp.h>
 
 struct socket tcb;
 
@@ -57,6 +52,8 @@ struct socket tcb;
 int	tcprexmtthresh = 3;
 struct	socket *tcp_last_so = &tcb;
 
+tcp_seq tcp_iss;                /* tcp initial send seq # */
+
 #define TCP_PAWS_IDLE	(24 * 24 * 60 * 60 * PR_SLOWHZ)
 
 /* for modulo comparisons of timestamps */
@@ -73,26 +70,52 @@ struct	socket *tcp_last_so = &tcb;
  * Set DELACK for segments received in order, but ack immediately
  * when segments are out of order (so fast retransmit can work).
  */
+#ifdef TCP_ACK_HACK
+#define        TCP_REASS(tp, ti, m, so, flags) {\
+       if ((ti)->ti_seq == (tp)->rcv_nxt && \
+           (tp)->seg_next == (tcpiphdrp_32)(tp) && \
+           (tp)->t_state == TCPS_ESTABLISHED) {\
+               if (ti->ti_flags & TH_PUSH) \
+                       tp->t_flags |= TF_ACKNOW; \
+               else \
+                       tp->t_flags |= TF_DELACK; \
+               (tp)->rcv_nxt += (ti)->ti_len; \
+               flags = (ti)->ti_flags & TH_FIN; \
+               tcpstat.tcps_rcvpack++;\
+               tcpstat.tcps_rcvbyte += (ti)->ti_len;\
+               if (so->so_emu) { \
+		       if (tcp_emu((so),(m))) \
+		       		sbappend((so), (m)); \
+	       } else \
+	       	       sbappend((so), (m)); \
+/*               sorwakeup(so); */ \
+	} else {\
+               (flags) = tcp_reass((tp), (ti), (m)); \
+               tp->t_flags |= TF_ACKNOW; \
+       } \
+}
+#else
 #define	TCP_REASS(tp, ti, m, so, flags) { \
 	if ((ti)->ti_seq == (tp)->rcv_nxt && \
-	    (tp)->seg_next == (struct tcpiphdr *)(tp) && \
+	    (tp)->seg_next == (tcpiphdrp_32)(tp) && \
 	    (tp)->t_state == TCPS_ESTABLISHED) { \
 		tp->t_flags |= TF_DELACK; \
 		(tp)->rcv_nxt += (ti)->ti_len; \
 		flags = (ti)->ti_flags & TH_FIN; \
 		tcpstat.tcps_rcvpack++;\
 		tcpstat.tcps_rcvbyte += (ti)->ti_len;\
-		if (so->so_iptos & IPTOS_EMU) { \
+		if (so->so_emu) { \
 			if (tcp_emu((so),(m))) \
 				sbappend(so, (m)); \
 		} else \
-			sbappend(so, (m)); \
+			sbappend((so), (m)); \
 /*		sorwakeup(so); */ \
 	} else { \
 		(flags) = tcp_reass((tp), (ti), (m)); \
 		tp->t_flags |= TF_ACKNOW; \
 	} \
 }
+#endif
 
 int
 tcp_reass(tp, ti, m)
@@ -103,15 +126,6 @@ tcp_reass(tp, ti, m)
 	register struct tcpiphdr *q;
 	struct socket *so = tp->t_socket;
 	int flags;
-
-	
-#ifdef DEBUG
-	debug_call(dfd,"tcp_reass(tp,ti,m) called ...\n");
-	debug_call(dfd,"    tp = %d\n    ti = %d\n    m = %d\n",
-		(int)tp,(int)ti,(int)m);
-	fflush_call(dfd);
-#endif
-	
 	
 	/*
 	 * Call with ti==0 after become established to
@@ -123,7 +137,7 @@ tcp_reass(tp, ti, m)
 	/*
 	 * Find a segment which begins after this one does.
 	 */
-	for (q = tp->seg_next; q != (struct tcpiphdr *)tp;
+	for (q = (struct tcpiphdr *)tp->seg_next; q != (struct tcpiphdr *)tp;
 	    q = (struct tcpiphdr *)q->ti_next)
 		if (SEQ_GT(q->ti_seq, ti->ti_seq))
 			break;
@@ -143,7 +157,13 @@ tcp_reass(tp, ti, m)
 				tcpstat.tcps_rcvduppack++;
 				tcpstat.tcps_rcvdupbyte += ti->ti_len;
 				m_freem(m);
-				return (0);
+				/*
+				 * Try to present any queued data
+				 * at the left window edge to the user.
+				 * This is needed after the 3-WHS
+				 * completes.
+				 */
+				goto present;   /* ??? */
 			}
 			m_adj(m, i);
 			ti->ti_len -= i;
@@ -153,7 +173,7 @@ tcp_reass(tp, ti, m)
 	}
 	tcpstat.tcps_rcvoopack++;
 	tcpstat.tcps_rcvoobyte += ti->ti_len;
-	REASS_MBUF(ti) = m;		/* XXX */
+	REASS_MBUF(ti) = (mbufp_32) m;		/* XXX */
 
 	/*
 	 * While we overlap succeeding segments trim them or,
@@ -166,28 +186,28 @@ tcp_reass(tp, ti, m)
 		if (i < q->ti_len) {
 			q->ti_seq += i;
 			q->ti_len -= i;
-			m_adj(REASS_MBUF(q), i);
+			m_adj((struct mbuf *) REASS_MBUF(q), i);
 			break;
 		}
 		q = (struct tcpiphdr *)q->ti_next;
-		m = REASS_MBUF((struct tcpiphdr *)q->ti_prev);
-		remque(q->ti_prev);
+		m = (struct mbuf *) REASS_MBUF((struct tcpiphdr *)q->ti_prev);
+		remque_32((void *)(q->ti_prev));
 		m_freem(m);
 	}
 
 	/*
 	 * Stick new segment in its place.
 	 */
-	insque(ti, q->ti_prev);
+	insque_32(ti, (void *)(q->ti_prev));
 
 present:
 	/*
 	 * Present data to user, advancing rcv_nxt through
 	 * completed sequence space.
 	 */
-	if (TCPS_HAVEESTABLISHED(tp->t_state) == 0)
+	if (!TCPS_HAVEESTABLISHED(tp->t_state))
 		return (0);
-	ti = tp->seg_next;
+	ti = (struct tcpiphdr *) tp->seg_next;
 	if (ti == (struct tcpiphdr *)tp || ti->ti_seq != tp->rcv_nxt)
 		return (0);
 	if (tp->t_state == TCPS_SYN_RECEIVED && ti->ti_len)
@@ -195,14 +215,14 @@ present:
 	do {
 		tp->rcv_nxt += ti->ti_len;
 		flags = ti->ti_flags & TH_FIN;
-		remque(ti);
-		m = REASS_MBUF(ti); /* XXX */
+		remque_32(ti);
+		m = (struct mbuf *) REASS_MBUF(ti); /* XXX */
 		ti = (struct tcpiphdr *)ti->ti_next;
 /*		if (so->so_state & SS_FCANTRCVMORE) */
 		if (so->so_state & SS_FCANTSENDMORE)
 			m_freem(m);
 		else {
-			if (so->so_iptos & IPTOS_EMU) {
+			if (so->so_emu) {
 				if (tcp_emu(so,m))
 					sbappend(so, m);
 			} else
@@ -237,11 +257,10 @@ tcp_input(m, iphlen, inso)
 	int ret;
 /*	int ts_present = 0; */
 
-#ifdef DEBUG
-	debug_call(dfd,"tcp_input(m,iphlen) called ...\n");
-	debug_call(dfd,"    m = %d\n    iphlen = %d\n", (int)m, iphlen);
-	fflush_call(dfd);
-#endif
+	DEBUG_CALL("tcp_input");
+	DEBUG_ARG("m = %lx", (long)m);
+	DEBUG_ARG("iphlen = %d", iphlen);
+	DEBUG_ARG("inso = %lx", (long)inso);
 	
 	/*
 	 * If called with m == 0, then we're continuing the connect
@@ -321,9 +340,7 @@ tcp_input(m, iphlen, inso)
  */
 	}
 	tiflags = ti->ti_flags;
-#ifdef DEBUG
-	debug_misc(dfd," --- tcp_input flags = %x\n", tiflags);
-#endif
+	
 	/*
 	 * Convert TCP protocol specific fields to host format.
 	 */
@@ -332,6 +349,12 @@ tcp_input(m, iphlen, inso)
 	NTOHS(ti->ti_win);
 	NTOHS(ti->ti_urp);
 
+	/*
+	 * Drop TCP, IP headers and TCP options.
+	 */
+	m->m_data += sizeof(struct tcpiphdr)+off-sizeof(struct tcphdr);
+	m->m_len  -= sizeof(struct tcpiphdr)+off-sizeof(struct tcphdr);
+	
 	/*
 	 * Locate pcb for segment.
 	 */
@@ -354,8 +377,8 @@ findso:
 	 * If the TCB exists but is in CLOSED state, it is embryonic,
 	 * but should either do a listen or a connect soon.
 	 *
-	 * state == CLOSED means we've done socreate() but havn't
-	 * attatched it to a protocol yet... 
+	 * state == CLOSED means we've done socreate() but haven't
+	 * attached it to a protocol yet... 
 	 * 
 	 * XXX If a TCB does not exist, and the TH_SYN flag is
 	 * the only flag set, then create a session, mark it
@@ -368,10 +391,13 @@ findso:
 		if ((so = socreate()) == NULL)
 		   goto dropwithreset;
 		if (tcp_attach(so) < 0) {
-			free(so); /* Not sofree, hasn't been insque'd yet */
+			free(so); /* Not sofree (if it failed, it's not insqued) */
 			goto dropwithreset;
 		}
 		
+		sbreserve(&so->so_snd, tcp_sndspace);
+		sbreserve(&so->so_rcv, tcp_rcvspace);
+		
 /*		tcp_last_so = so; */  /* XXX ? */
 		tp = sototcpcb(so);
 		
@@ -389,7 +415,7 @@ findso:
            
         /*
          * If this is a still-connecting socket, this probably a
-         * a retransmit of the SYN.  Wheather it's a retrasmit SYN
+         * a retransmit of the SYN.  Whether it's a retransmit SYN
 	 * or something else, we nuke it.
          */
         if (so->so_state & SS_ISFCONNECTING)
@@ -463,7 +489,7 @@ findso:
 		if (ti->ti_len == 0) {
 			if (SEQ_GT(ti->ti_ack, tp->snd_una) &&
 			    SEQ_LEQ(ti->ti_ack, tp->snd_max) &&
-			    tp->snd_cwnd >= tp->snd_wnd ) {
+			    tp->snd_cwnd >= tp->snd_wnd) {
 				/*
 				 * this is a pure ack for outstanding data.
 				 */
@@ -471,8 +497,7 @@ findso:
 /*				if (ts_present)
  *					tcp_xmit_timer(tp, tcp_now-ts_ecr+1);
  *				else 
- */				     
-				     if (tp->t_rtt &&
+ */				     if (tp->t_rtt &&
 					    SEQ_GT(ti->ti_ack, tp->t_rtseq))
 					tcp_xmit_timer(tp, tp->t_rtt);
 				acked = ti->ti_ack - tp->snd_una;
@@ -506,18 +531,15 @@ findso:
 				/* 
 				 * This is called because sowwakeup might have
 				 * put data into so_snd.  Since we don't so sowwakeup,
-				 * we don't need this..
+				 * we don't need this.. XXX???
 				 */
-/*				if (so->so_snd.sb_cc)
- *					(void) tcp_output(tp);
- */
-#ifdef DEBUG
-				debug_misc(dfd,"returning 1\n");
-#endif
+				if (so->so_snd.sb_cc)
+					(void) tcp_output(tp);
+
 				return;
 			}
 		} else if (ti->ti_ack == tp->snd_una &&
-		    tp->seg_next == (struct tcpiphdr *)tp &&
+		    tp->seg_next == (tcpiphdrp_32)tp &&
 		    ti->ti_len <= sbspace(&so->so_rcv)) {
 			/*
 			 * this is a pure, in-sequence data packet
@@ -529,12 +551,9 @@ findso:
 			tcpstat.tcps_rcvpack++;
 			tcpstat.tcps_rcvbyte += ti->ti_len;
 			/*
-			 * Drop TCP, IP headers and TCP options then add data
-			 * to socket buffer.
+			 * Add data to socket buffer.
 			 */
-			m->m_data += sizeof(struct tcpiphdr)+off-sizeof(struct tcphdr);
-			m->m_len -= sizeof(struct tcpiphdr)+off-sizeof(struct tcphdr);
-			if (so->so_iptos & IPTOS_EMU) {
+			if (so->so_emu) {
 				if (tcp_emu(so,m))
 					sbappend(so, m);
 			} else
@@ -546,13 +565,13 @@ findso:
 			 * XXX Need to check? It's be NON_BLOCKING
 			 */
 /*			sorwakeup(so); */
-
+			
 			/*
-			 * If this is a small packet, then ACK now - with Nagel
+			 * If this is a short packet, then ACK now - with Nagel
 			 *	congestion avoidance sender won't send more until
 			 *	he gets an ACK.
 			 * 
-			 * Here are 3 interpritations of what should happend.
+			 * Here are 3 interpretations of what should happen.
 			 * The best (for me) is to delay-ack everything except
 			 * if it's a one-byte packet containing an ESC
 			 * (this means it's an arrow key (or similar) sent using
@@ -570,25 +589,13 @@ findso:
 			if ((unsigned)ti->ti_len == 1 &&
 			    ((struct tcpiphdr_2 *)ti)->first_char == (char)27) {
 				tp->t_flags |= TF_ACKNOW;
-#ifdef DEBUG
-				debug_misc(dfd," tcp_output called from tcp_input 1\n");
-#endif
 				tcp_output(tp);
 			} else {
 				tp->t_flags |= TF_DELACK;
 			}
-#ifdef DEBUG
-			debug_misc(dfd," returning 2\n");
-#endif
 			return;
 		}
 	}
-	/*
-	 * Drop TCP, IP headers and TCP options.
-	 */
-	m->m_data += sizeof(struct tcpiphdr)+off-sizeof(struct tcphdr);
-	m->m_len  -= sizeof(struct tcpiphdr)+off-sizeof(struct tcphdr);
-
 	/*
 	 * Calculate amount of space in receive window,
 	 * and then do TCP input processing.
@@ -636,37 +643,32 @@ findso:
 		 * If this is destined for the control address, then flag to
 		 * tcp_ctl once connected, otherwise connect
 		 */
-		if ((ntohl(so->so_faddr.s_addr) & 0xffffff00) == ntohl(special_addr)) {
-			if ((ntohl(so->so_faddr.s_addr) & 0xff) == CTL_EXEC) {
-				struct ex_list *ex_ptr;
-				
-				/* Check if there's something to exec for this session */
-				for (ex_ptr = exec_list; ex_ptr; ex_ptr = ex_ptr->ex_next)  {
-					if (ex_ptr->ex_fport == so->so_fport &&
-					    (ntohl(so->so_faddr.s_addr) & 0xff) == ex_ptr->ex_addr) {
-						so->so_state |= SS_CTL;
-						goto cont_input;
-					}
-				}
-				/* Else, just run tcp_fconnect */
-			} else if ((ntohl(so->so_faddr.s_addr) & 0xff) == CTL_ALIAS) {
-				/* Do nothing */
+		if ((so->so_faddr.s_addr & ntohl(0xffffff00)) == special_addr.s_addr) {
+			if ((ntohl(so->so_faddr.s_addr) & 0xff) == CTL_ALIAS) {
+				/* Do nothing, tcp_fconnect will be called on it */
 			} else {
+				/* Must be an add exec */
 				so->so_state |= SS_CTL;
 				goto cont_input;
 			}
 		}
-
+		
+		if (so->so_emu & EMU_NOCONNECT) {
+			so->so_emu &= ~EMU_NOCONNECT;
+			goto cont_input;
+		}
+		
 		tcp_fconnect(so);
 		/*
-		 * Havnt connected yet, save the current mbuf
+		 * Haven't connected yet, save the current mbuf
 		 * and ti, and return
-		 * XXX Some OS's don't tell us wheather the connect()
+		 * XXX Some OS's don't tell us whether the connect()
 		 * succeeded or not.  So we must time it out.
 		 */
 		so->so_m = m;
 		so->so_ti = ti;
 		tp->t_timer[TCPT_KEEP] = TCPTV_KEEP_INIT;
+		tp->t_state = TCPS_SYN_RECEIVED;
 		return;
 
 cont_conn:
@@ -825,36 +827,26 @@ trimthenstep6:
 				tiflags &= ~TH_URG;
 			todrop--;
 		}
-		if (todrop >= ti->ti_len) {
-			tcpstat.tcps_rcvduppack++;
-			tcpstat.tcps_rcvdupbyte += ti->ti_len;
+		/*
+		 * Following if statement from Stevens, vol. 2, p. 960.
+		 */
+		if (todrop > ti->ti_len
+		    || (todrop == ti->ti_len && (tiflags & TH_FIN) == 0)) {
 			/*
-			 * If segment is just one to the left of the window,
-			 * check two special cases:
-			 * 1. Don't toss RST in response to 4.2-style keepalive.
-			 * 2. If the only thing to drop is a FIN, we can drop
-			 *    it, but check the ACK or we will get into FIN
-			 *    wars if our FINs crossed (both CLOSING).
-			 * In either case, send ACK to resynchronize,
-			 * but keep on processing for RST or ACK.
+			 * Any valid FIN must be to the left of the window.
+			 * At this point the FIN must be a duplicate or out
+			 * of sequence; drop it.
 			 */
-			if ((tiflags & TH_FIN && todrop == ti->ti_len + 1)
-#ifdef TCP_COMPAT_42
-			  || (tiflags & TH_RST && ti->ti_seq == tp->rcv_nxt - 1)
-#endif
-			   ) {
-				todrop = ti->ti_len;
-				tiflags &= ~TH_FIN;
-				tp->t_flags |= TF_ACKNOW;
-			} else {
-				/*
-				 * Handle the case when a bound socket connects
-				 * to itself. Allow packets with a SYN and
-				 * an ACK to continue with the processing.
-				 */
-				if (todrop != 0 || (tiflags & TH_ACK) == 0)
-					goto dropafterack;
-			}
+			tiflags &= ~TH_FIN;
+			
+			/*
+			 * Send an ACK to resynchronize and drop any data.
+			 * But keep on processing for RST or ACK.
+			 */
+			tp->t_flags |= TF_ACKNOW;
+			todrop = ti->ti_len;
+			tcpstat.tcps_rcvduppack++;
+			tcpstat.tcps_rcvdupbyte += todrop;
 		} else {
 			tcpstat.tcps_rcvpartduppack++;
 			tcpstat.tcps_rcvpartdupbyte += todrop;
@@ -1002,7 +994,7 @@ trimthenstep6:
 			/* So tcp_ctl reports the right state */
 			tp->t_state = TCPS_ESTABLISHED;
 			ret = tcp_ctl(so);
-			if (ret == 1)  {
+			if (ret == 1) {
 				soisfconnected(so);
 				so->so_state &= ~SS_CTL; /* XXX */
 			} else if (ret == 2) {
@@ -1052,7 +1044,7 @@ trimthenstep6:
 				 * duplicate ack (ie, window info didn't
 				 * change), the ack is the biggest we've
 				 * seen and we've seen exactly our rexmt
-				 * threshhold of them, assume a packet
+				 * threshold of them, assume a packet
 				 * has been dropped and retransmit it.
 				 * Kludge snd_nxt & the congestion
 				 * window so we send only this one
@@ -1086,9 +1078,6 @@ trimthenstep6:
 					tp->t_rtt = 0;
 					tp->snd_nxt = ti->ti_ack;
 					tp->snd_cwnd = tp->t_maxseg;
-#ifdef DEBUG
-					debug_misc(dfd," tcp_output called from tcp_input 2\n");
-#endif
 					(void) tcp_output(tp);
 					tp->snd_cwnd = tp->snd_ssthresh +
 					       tp->t_maxseg * tp->t_dupacks;
@@ -1097,9 +1086,6 @@ trimthenstep6:
 					goto drop;
 				} else if (tp->t_dupacks > tcprexmtthresh) {
 					tp->snd_cwnd += tp->t_maxseg;
-#ifdef DEBUG
-					debug_misc(dfd," tcp_ouput called from tcp_input 3\n");
-#endif
 					(void) tcp_output(tp);
 					goto drop;
 				}
@@ -1148,9 +1134,6 @@ trimthenstep6:
 		if (ti->ti_ack == tp->snd_max) {
 			tp->t_timer[TCPT_REXMT] = 0;
 			needoutput = 1;
-#ifdef DEBUG
-			debug_misc(dfd," needed output 1\n");
-#endif
 		} else if (tp->t_timer[TCPT_PERSIST] == 0)
 			tp->t_timer[TCPT_REXMT] = tp->t_rxtcur;
 		/*
@@ -1179,7 +1162,7 @@ trimthenstep6:
 		}
 		/*
 		 * XXX sowwakup is called when data is acked and there's room for
-		 * for more data... it shoudl read() the socket 
+		 * for more data... it should read() the socket 
 		 */
 /*		if (so->so_snd.sb_flags & SB_NOTIFY)
  *			sowwakeup(so);
@@ -1270,9 +1253,6 @@ step6:
 		if (tp->snd_wnd > tp->max_sndwnd)
 			tp->max_sndwnd = tp->snd_wnd;
 		needoutput = 1;
-#ifdef DEBUG
-		debug_misc(dfd," needed output 2\n");
-#endif
 	}
 
 	/*
@@ -1309,9 +1289,6 @@ step6:
 			tp->rcv_up = ti->ti_seq + ti->ti_urp;
 			so->so_urgc =  so->so_rcv.sb_cc +
 				(tp->rcv_up - tp->rcv_nxt); /* -1; */
-#ifdef DEBUG
-			debug_misc(dfd,"  --- we have %d bytes of urgent data\n", so->so_urgc);
-#endif
 			tp->rcv_up = ti->ti_seq + ti->ti_urp;
 	 
 		}
@@ -1415,10 +1392,10 @@ dodata:
 /*	if ((ti->ti_len && (unsigned)ti->ti_len < tp->t_maxseg &&
  *		(so->so_iptos & IPTOS_LOWDELAY) == 0) ||
  *	       ((so->so_iptos & IPTOS_LOWDELAY) &&
- *	       ((struct tcpiphdr_2 *)ti)->first_char == (char)27))  {
+ *	       ((struct tcpiphdr_2 *)ti)->first_char == (char)27)) {
  */
 	if (ti->ti_len && (unsigned)ti->ti_len <= 5 &&
-	    ((struct tcpiphdr_2 *)ti)->first_char == (char)27)  {
+	    ((struct tcpiphdr_2 *)ti)->first_char == (char)27) {
 		tp->t_flags |= TF_ACKNOW;
 	}
 
@@ -1426,15 +1403,8 @@ dodata:
 	 * Return any desired output.
 	 */
 	if (needoutput || (tp->t_flags & TF_ACKNOW)) {
-#ifdef DEBUG
-		debug_misc(dfd," tp->t_flags == %x, tp->t_state = %x\n", tp->t_flags, tp->t_state);
-		debug_misc(dfd," tcp_output called from tcp_input 5\n");
-#endif
 		(void) tcp_output(tp);
 	}
-#ifdef DEBUG
-	debug_misc(dfd," returning  5\n");
-#endif
 	return;
 
 dropafterack:
@@ -1446,9 +1416,6 @@ dropafterack:
 		goto drop;
 	m_freem(m);
 	tp->t_flags |= TF_ACKNOW;
-#ifdef DEBUG
-	debug_misc(dfd," tcp_output called from tcp_input 6\n");
-#endif
 	(void) tcp_output(tp);
 	return;
 
@@ -1473,15 +1440,16 @@ drop:
 	return;
 }
 
+ /* , ts_present, ts_val, ts_ecr) */
+/*	int *ts_present;
+ *	u_int32_t *ts_val, *ts_ecr;
+ */
 void
-tcp_dooptions(tp, cp, cnt, ti) /* , ts_present, ts_val, ts_ecr) */
+tcp_dooptions(tp, cp, cnt, ti)
 	struct tcpcb *tp;
 	u_char *cp;
 	int cnt;
 	struct tcpiphdr *ti;
-/*	int *ts_present;
- *	u_int32_t *ts_val, *ts_ecr;
- */
 {
 	u_int16_t mss;
 	int opt, optlen;
@@ -1595,13 +1563,9 @@ tcp_xmit_timer(tp, rtt)
 {
 	register short delta;
 
-	
-#ifdef DEBUG
-	debug_call(dfd,"tcp_xmit_timer(tp,rtt) called ...\n");
-	debug_call(dfd,"    tp = %d\n    rtt = %d\n", (int)tp,rtt);
-	fflush_call(dfd);
-#endif
-	
+	DEBUG_CALL("tcp_xmit_timer");
+	DEBUG_ARG("tp = %lx", (long)tp);
+	DEBUG_ARG("rtt = %d", rtt);
 	
 	tcpstat.tcps_rttupdated++;
 	if (tp->t_srtt != 0) {
@@ -1654,7 +1618,7 @@ tcp_xmit_timer(tp, rtt)
 	 * the minimum feasible timer (which is 2 ticks).
 	 */
 	TCPT_RANGESET(tp->t_rxtcur, TCP_REXMTVAL(tp),
-	    tp->t_rttmin, TCPTV_REXMTMAX);
+	    (short)tp->t_rttmin, TCPTV_REXMTMAX); /* XXX */
 	
 	/*
 	 * We received an ack for a packet that wasn't retransmitted;
@@ -1682,169 +1646,31 @@ tcp_xmit_timer(tp, rtt)
  * parameters from pre-set or cached values in the routing entry.
  */
 
-/*
- * XXX Needs work
- */
 int
 tcp_mss(tp, offer)
         register struct tcpcb *tp;
         u_int offer;
 {
+	struct socket *so = tp->t_socket;
 	int mss;
 	
-#ifdef DEBUG
-	debug_call(dfd,"tcp_mss(tp,offer) called ...\n");
-	debug_call(dfd,"    tp = %d\n    offer = %d\n", (int)tp,offer);
-	fflush_call(dfd);
-#endif
-	mss = ifp.ifp_mtu - sizeof(struct tcpiphdr);
+	DEBUG_CALL("tcp_mss");
+	DEBUG_ARG("tp = %lx", (long)tp);
+	DEBUG_ARG("offer = %d", offer);
+	
+	mss = min(if_mtu, if_mru) - sizeof(struct tcpiphdr);
 	if (offer)
 		mss = min(mss, offer);
 	mss = max(mss, 32);
-	if (mss < tp->t_maxseg || offer != 0) {
-		tp->t_maxseg = mss;
-	}
-		
+	if (mss < tp->t_maxseg || offer != 0)
+	   tp->t_maxseg = mss;
+	
 	tp->snd_cwnd = mss;
+	
+	sbreserve(&so->so_snd, tcp_sndspace+((tcp_sndspace%mss)?(mss-(tcp_sndspace%mss)):0));
+	sbreserve(&so->so_rcv, tcp_rcvspace+((tcp_rcvspace%mss)?(mss-(tcp_rcvspace%mss)):0));
+	
+	DEBUG_MISC((dfd, "returning mss = %d\n", mss));
+	
 	return mss;
 }
-
-#ifdef notdef
-
-int
-tcp_mss(tp, offer)
-	register struct tcpcb *tp;
-	u_int offer;
-{
-	struct route *ro;
-	register struct rtentry *rt;
-	struct ifnet *ifp;
-	register int rtt, mss;
-	u_long bufsize;
-	struct inpcb *inp;
-	struct socket *so;
-	extern int tcp_mssdflt;
-
-	inp = tp->t_inpcb;
-	ro = &inp->inp_route;
-
-	if ((rt = ro->ro_rt) == (struct rtentry *)0) {
-		/* No route yet, so try to acquire one */
-		if (inp->inp_faddr.s_addr != INADDR_ANY) {
-			ro->ro_dst.sa_family = AF_INET;
-			ro->ro_dst.sa_len = sizeof(ro->ro_dst);
-			((struct sockaddr_in *) &ro->ro_dst)->sin_addr =
-				inp->inp_faddr;
-			rtalloc(ro);
-		}
-		if ((rt = ro->ro_rt) == (struct rtentry *)0)
-			return (tcp_mssdflt);
-	}
-	ifp = rt->rt_ifp;
-	so = inp->inp_socket;
-
-#ifdef RTV_MTU	/* if route characteristics exist ... */
-	/*
-	 * While we're here, check if there's an initial rtt
-	 * or rttvar.  Convert from the route-table units
-	 * to scaled multiples of the slow timeout timer.
-	 */
-	if (tp->t_srtt == 0 && (rtt = rt->rt_rmx.rmx_rtt)) {
-		/*
-		 * XXX the lock bit for MTU indicates that the value
-		 * is also a minimum value; this is subject to time.
-		 */
-		if (rt->rt_rmx.rmx_locks & RTV_RTT)
-			tp->t_rttmin = rtt / (RTM_RTTUNIT / PR_SLOWHZ);
-		tp->t_srtt = rtt / (RTM_RTTUNIT / (PR_SLOWHZ * TCP_RTT_SCALE));
-		if (rt->rt_rmx.rmx_rttvar)
-			tp->t_rttvar = rt->rt_rmx.rmx_rttvar /
-			    (RTM_RTTUNIT / (PR_SLOWHZ * TCP_RTTVAR_SCALE));
-		else
-			/* default variation is +- 1 rtt */
-			tp->t_rttvar =
-			    tp->t_srtt * TCP_RTTVAR_SCALE / TCP_RTT_SCALE;
-		TCPT_RANGESET(tp->t_rxtcur,
-		    ((tp->t_srtt >> 2) + tp->t_rttvar) >> 1,
-		    tp->t_rttmin, TCPTV_REXMTMAX);
-	}
-	/*
-	 * if there's an mtu associated with the route, use it
-	 */
-	if (rt->rt_rmx.rmx_mtu)
-		mss = rt->rt_rmx.rmx_mtu - sizeof(struct tcpiphdr);
-	else
-#endif /* RTV_MTU */
-	{
-		mss = ifp->if_mtu - sizeof(struct tcpiphdr);
-#if	(MCLBYTES & (MCLBYTES - 1)) == 0
-		if (mss > MCLBYTES)
-			mss &= ~(MCLBYTES-1);
-#else
-		if (mss > MCLBYTES)
-			mss = mss / MCLBYTES * MCLBYTES;
-#endif
-		if (!in_localaddr(inp->inp_faddr))
-			mss = min(mss, tcp_mssdflt);
-	}
-	/*
-	 * The current mss, t_maxseg, is initialized to the default value.
-	 * If we compute a smaller value, reduce the current mss.
-	 * If we compute a larger value, return it for use in sending
-	 * a max seg size option, but don't store it for use
-	 * unless we received an offer at least that large from peer.
-	 * However, do not accept offers under 32 bytes.
-	 */
-	if (offer)
-		mss = min(mss, offer);
-	mss = max(mss, 32);		/* sanity */
-	if (mss < tp->t_maxseg || offer != 0) {
-		/*
-		 * If there's a pipesize, change the socket buffer
-		 * to that size.  Make the socket buffers an integral
-		 * number of mss units; if the mss is larger than
-		 * the socket buffer, decrease the mss.
-		 */
-#ifdef RTV_SPIPE
-		if ((bufsize = rt->rt_rmx.rmx_sendpipe) == 0)
-#endif
-			bufsize = so->so_snd.sb_datalen;
-		if (bufsize < mss)
-			mss = bufsize;
-		else {
-			bufsize = roundup(bufsize, mss);
-			if (bufsize > sb_max)
-				bufsize = sb_max;
-			(void)sbreserve(&so->so_snd, bufsize);
-		}
-		tp->t_maxseg = mss;
-
-#ifdef RTV_RPIPE
-		if ((bufsize = rt->rt_rmx.rmx_recvpipe) == 0)
-#endif
-			bufsize = so->so_rcv.sb_datalen;
-		if (bufsize > mss) {
-			bufsize = roundup(bufsize, mss);
-			if (bufsize > sb_max)
-				bufsize = sb_max;
-			(void)sbreserve(&so->so_rcv, bufsize);
-		}
-	}
-	tp->snd_cwnd = mss;
-
-#ifdef RTV_SSTHRESH
-	if (rt->rt_rmx.rmx_ssthresh) {
-		/*
-		 * There's some sort of gateway or interface
-		 * buffer limit on the path.  Use this to set
-		 * the slow start threshhold, but set the
-		 * threshold to no less than 2*mss.
-		 */
-		tp->snd_ssthresh = max(2 * mss, rt->rt_rmx.rmx_ssthresh);
-	}
-#endif /* RTV_MTU */
-	return (mss);
-}
-
-#endif /* notdef */
-
