@@ -42,21 +42,23 @@
  * terms and conditions of the copyright.
  */
 
-#include "h/common.h"
-#include "h/ip.h"
-#include "h/tcp.h"
-#include "h/socket.h"
+#include <slirp.h>
 
 #define max(x,y) ((x) > (y) ? (x) : (y))
 #define min(x,y) ((x) < (y) ? (x) : (y))
 
-char *tcpstates[] =  {
-	"CLOSED",       "LISTEN",       "SYN_SENT",     "SYN_RCVD",
+/*
+ * Since this is only used in "stats socket", we give meaning
+ * names instead of the REAL names
+ */
+char *tcpstates[] = {
+/*	"CLOSED",       "LISTEN",       "SYN_SENT",     "SYN_RCVD", */
+	"REDIRECT",	"LISTEN",	"SYN_SENT",     "SYN_RCVD",
 	"ESTABLISHED",  "CLOSE_WAIT",   "FIN_WAIT_1",   "CLOSING",
 	"LAST_ACK",     "FIN_WAIT_2",   "TIME_WAIT",
 };
 
-u_char  tcp_outflags[TCP_NSTATES] =  {
+u_char  tcp_outflags[TCP_NSTATES] = {
 	TH_RST|TH_ACK, 0, TH_SYN, TH_SYN|TH_ACK,
 	TH_ACK, TH_ACK,
 	TH_FIN|TH_ACK, TH_FIN|TH_ACK, TH_FIN|TH_ACK, TH_ACK, TH_ACK,
@@ -80,15 +82,9 @@ tcp_output(tp)
 	u_char opt[MAX_TCPOPTLEN];
 	unsigned optlen, hdrlen;
 	int idle, sendalot;
-
-#ifdef DEBUG
-	debug_call(dfd,"tcp_output(tp) called ...\n");
-	debug_call(dfd,"    tp = %d\n", (int)tp);
-	fflush_call(dfd);
-	debug_call(dfd,"    so->so_snd.sb_cc = %d\n", so->so_snd.sb_cc);
-	debug_call(dfd,"    tp->t_state = %d\n", tp->t_state);
-	fflush_call(dfd);
-#endif
+	
+	DEBUG_CALL("tcp_output");
+	DEBUG_ARG("tp = %lx", (long)tp);
 	
 	/*
 	 * Determine length of data that should be transmitted,
@@ -110,9 +106,9 @@ again:
 	win = min(tp->snd_wnd, tp->snd_cwnd);
 
 	flags = tcp_outflags[tp->t_state];
-#ifdef DEBUG
-	debug_misc(dfd," --- tcp_output flags = %x\n",flags);
-#endif
+	
+	DEBUG_MISC((dfd, " --- tcp_output flags = %x\n",flags));
+	
 	/*
 	 * If in persist timeout with window of 0, send 1 byte.
 	 * Otherwise, if window is small but nonzero
@@ -165,6 +161,7 @@ again:
 			tp->snd_nxt = tp->snd_una;
 		}
 	}
+	
 	if (len > tp->t_maxseg) {
 		len = tp->t_maxseg;
 		sendalot = 1;
@@ -192,7 +189,7 @@ again:
 			goto send;
 		if (tp->t_force)
 			goto send;
-		if (len >= tp->max_sndwnd / 2)
+		if (len >= tp->max_sndwnd / 2 && tp->max_sndwnd > 0)
 			goto send;
 		if (SEQ_LT(tp->snd_nxt, tp->snd_max))
 			goto send;
@@ -269,9 +266,6 @@ again:
 	/*
 	 * No reason to send a segment, just return.
 	 */
-#ifdef DEBUG
-	debug_misc(dfd,"tcp_output: Didn't do anything\n");
-#endif
 	tcpstat.tcps_didnuttin++;
 	
 	return (0);
@@ -363,7 +357,7 @@ send:
 			error = 1;
 			goto out;
 		}
-		m->m_data += ifp.ifp_maxlinkhdr;
+		m->m_data += if_maxlinkhdr;
 		m->m_len = hdrlen;
 		
 		/* 
@@ -405,7 +399,7 @@ send:
 			error = 1;
 			goto out;
 		}
-		m->m_data += ifp.ifp_maxlinkhdr;
+		m->m_data += if_maxlinkhdr;
 		m->m_len = hdrlen;
 	}
 
@@ -542,7 +536,7 @@ send:
 	((struct ip *)ti)->ip_len = m->m_len;
 
 	((struct ip *)ti)->ip_ttl = ip_defttl;
-	((struct ip *)ti)->ip_tos = (so->so_iptos & ~IPTOS_EMU);
+	((struct ip *)ti)->ip_tos = so->so_iptos;
 	    
 /* #if BSD >= 43 */
 	/* Don't do IP options... */
@@ -600,7 +594,7 @@ tcp_setpersist(tp)
  *		panic("tcp_output REXMT");
  */
 	/*
-	 * Start/restart persistance timer.
+	 * Start/restart persistence timer.
 	 */
 	TCPT_RANGESET(tp->t_timer[TCPT_PERSIST],
 	    t * tcp_backoff[tp->t_rxtshift],
