@@ -42,29 +42,13 @@
  * terms and conditions of the copyright.
  */
 
-#include "h/common.h"
-#include <sys/socket.h>
-
-#ifdef HAVE_SYS_FILIO_H
-#include <sys/filio.h>
-#endif
-
-#ifdef HAVE_SYS_IOCTL_H
-#include <sys/ioctl.h>
-#endif
-
-#include "h/ip.h"
-#include "h/tcp.h"
-#include "h/udp.h"
-#include "h/socket.h"
-#include "h/main.h"
-#include "h/ctl.h"
-#include <string.h>
+#define WANT_SYS_IOCTL_H
+#include <slirp.h>
 
 /* patchable/settable parameters for tcp */
 int 	tcp_mssdflt = TCP_MSS;
 int 	tcp_rttdflt = TCPTV_SRTTDFLT / PR_SLOWHZ;
-int	tcp_do_rfc1323 = 0;	/* Don't do rfc1323 preformance enhancments */
+int	tcp_do_rfc1323 = 0;	/* Don't do rfc1323 performance enhancements */
 int	tcp_rcvspace;	/* You may want to change this */
 int	tcp_sndspace;	/* Keep small if you have an error prone link */
 
@@ -77,12 +61,13 @@ tcp_init()
 	tcp_iss = 1;		/* wrong */
 	tcb.so_next = tcb.so_prev = &tcb;
 	
+	/* tcp_rcvspace = our Window we advertise to the remote */
 	tcp_rcvspace = TCP_RCVSPACE;
 	tcp_sndspace = TCP_SNDSPACE;
 	
-	/* Make sure tcp_sndspace is at *least* 2*MSS */
-	if (tcp_sndspace < 2*(ifp.ifp_mtu - sizeof(struct tcpiphdr)))
-		tcp_sndspace = 2*(ifp.ifp_mtu - sizeof(struct tcpiphdr));
+	/* Make sure tcp_sndspace is at least 2*MSS */
+	if (tcp_sndspace < 2*(min(if_mtu, if_mru) - sizeof(struct tcpiphdr)))
+		tcp_sndspace = 2*(min(if_mtu, if_mru) - sizeof(struct tcpiphdr));
 }
 
 /*
@@ -99,12 +84,6 @@ tcp_template(tp)
 	struct socket *so = tp->t_socket;
 	register struct tcpiphdr *n = &tp->t_template;
 
-#ifdef DEBUG
-	debug_call(dfd,"tcp_template(tp) called ...\n");
-	debug_call(dfd,"    tp = %d\n", (int)tp);
-	fflush_call(dfd);
-#endif
-	
 	n->ti_next = n->ti_prev = 0;
 	n->ti_x1 = 0;
 	n->ti_pr = IPPROTO_TCP;
@@ -148,13 +127,13 @@ tcp_respond(tp, ti, m, ack, seq, flags)
 	register int tlen;
 	int win = 0;
 
-#ifdef DEBUG
-	debug_call(dfd,"tcp_respond(tp,ti,m,ack,seq,flags) called ...\n");
-	debug_call(dfd,"    tp = %d\n    ti = %d\n    m = %d\n    ack = %d\n\
-    seq = %d\n    flags = %d\n", (int)tp, (int)ti, (int)m, (int)ack, (int)seq, flags);
-	fflush_call(dfd);
-#endif
-	
+	DEBUG_CALL("tcp_respond");
+	DEBUG_ARG("tp = %lx", (long)tp);
+	DEBUG_ARG("ti = %lx", (long)ti);
+	DEBUG_ARG("m = %lx", (long)m);
+	DEBUG_ARG("ack = %d", ack);
+	DEBUG_ARG("seq = %d", seq);
+	DEBUG_ARG("flags = %x", flags);
 	
 	if (tp)
 		win = sbspace(&tp->t_socket->so_rcv);
@@ -166,7 +145,7 @@ tcp_respond(tp, ti, m, ack, seq, flags)
 #else
 		tlen = 0;
 #endif
-		m->m_data += ifp.ifp_maxlinkhdr;
+		m->m_data += if_maxlinkhdr;
 		*mtod(m, struct tcpiphdr *) = *ti;
 		ti = mtod(m, struct tcpiphdr *);
 		flags = TH_ACK;
@@ -220,20 +199,14 @@ tcp_newtcpcb(so)
 {
 	register struct tcpcb *tp;
 	
-#ifdef DEBUG
-	debug_call(dfd,"tcp_newtcpcb(so) called ...\n");
-	debug_call(dfd,"    so = %d\n", (int)so);
-	fflush_call(dfd);
-#endif
-	
 	tp = (struct tcpcb *)malloc(sizeof(*tp));
 	if (tp == NULL)
 		return ((struct tcpcb *)0);
 	
 	memset((char *) tp, 0, sizeof(struct tcpcb));
-	tp->seg_next = tp->seg_prev = (struct tcpiphdr *)tp;
-	tp->t_maxseg = tcp_mssdflt; /* XXX */
-
+	tp->seg_next = tp->seg_prev = (tcpiphdrp_32)tp;
+	tp->t_maxseg = tcp_mssdflt;
+	
 	tp->t_flags = tcp_do_rfc1323 ? (TF_REQ_SCALE|TF_REQ_TSTMP) : 0;
 	tp->t_socket = so;
 	
@@ -270,18 +243,12 @@ tcp_drop(tp, errno)
 	int errno;
 {
 
-#ifdef DEBUG
-	debug_call(dfd,"tcp_drop(tp,errno) called ...\n");
-	debug_call(dfd,"    tp = %d\n    errno = %d\n", (int)tp, errno);
-	fflush_call(dfd);
-#endif
-	
+	DEBUG_CALL("tcp_drop");
+	DEBUG_ARG("tp = %lx", (long)tp);
+	DEBUG_ARG("errno = %d", errno);
 	
 	if (TCPS_HAVERCVDSYN(tp->t_state)) {
 		tp->t_state = TCPS_CLOSED;
-#ifdef DEBUG
-		debug_misc(dfd," tcp_output  called  from tcp_drop\n");
-#endif
 		(void) tcp_output(tp);
 		tcpstat.tcps_drops++;
 	} else
@@ -306,20 +273,13 @@ tcp_close(tp)
 	register struct tcpiphdr *t;
 	struct socket *so = tp->t_socket;
 	register struct mbuf *m;
-
-#ifdef DEBUG
-	debug_call(dfd,"tcp_close(tp) called ...\n");
-	fflush_call(dfd);
-	debug_call(dfd,"    tp = %d\n", (int)tp);
-	fflush_call(dfd);
-#endif
-
+	
 	/* free the reassembly queue, if any */
-	t = tp->seg_next;
+	t = (struct tcpiphdr *) tp->seg_next;
 	while (t != (struct tcpiphdr *)tp) {
 		t = (struct tcpiphdr *)t->ti_next;
-		m = REASS_MBUF((struct tcpiphdr *)t->ti_prev);
-		remque(t->ti_prev);
+		m = (struct mbuf *) REASS_MBUF((struct tcpiphdr *)t->ti_prev);
+		remque_32((struct tcpiphdr *) t->ti_prev);
 		m_freem(m);
 	}
 	/* It's static */
@@ -386,11 +346,8 @@ tcp_sockclosed(tp)
 	struct tcpcb *tp;
 {
 
-#ifdef DEBUG
-	debug_call(dfd,"tcp_sockclosed(tp) called ...\n");
-	debug_call(dfd,"    tp = %d\n", (int)tp);
-	fflush_call(dfd);
-#endif
+	DEBUG_CALL("tcp_sockclosed");
+	DEBUG_ARG("tp = %lx", (long)tp);
 	
 	switch (tp->t_state) {
 
@@ -420,7 +377,7 @@ tcp_sockclosed(tp)
 /* 
  * Connect to a host on the Internet
  * Called by tcp_input
- * Only do a connect, the tcp feilds will be set in tcp_input
+ * Only do a connect, the tcp fields will be set in tcp_input
  * return 0 if there's a result of the connect,
  * else return -1 means we're still connecting
  */
@@ -432,11 +389,8 @@ tcp_fconnect(so)
 	int s, ret;
 	int opt = 1;
 	
-#ifdef DEBUG
-	debug_call(dfd,"tcp_fconnect(so) called ...\n");
-	debug_call(dfd,"    so = %d\n", (int)so);
-	fflush_call(dfd);
-#endif
+	DEBUG_CALL("tcp_fconnect");
+	DEBUG_ARG("so = %lx", (long)so);
 
 	s = socket(AF_INET,SOCK_STREAM,0);
 	if (s < 0)
@@ -446,9 +400,18 @@ tcp_fconnect(so)
 	setsockopt(s,SOL_SOCKET,SO_OOBINLINE,(char *)&opt,sizeof(int));
 	
 	addr.sin_family = AF_INET;
-	if ((ntohl(so->so_faddr.s_addr) & 0xffffff00) == ntohl(special_addr))
-		addr.sin_addr.s_addr = inet_addr("127.0.0.1");
-	else
+	if ((so->so_faddr.s_addr & htonl(0xffffff00)) == special_addr.s_addr) {
+		/* It's an alias */
+		switch(ntohl(so->so_faddr.s_addr) & 0xff) {
+		 case CTL_DNS:
+			addr.sin_addr = dns_addr;
+			break;
+		 case CTL_ALIAS:
+		 default:
+			addr.sin_addr = loopback_addr;
+			break;
+		}
+	} else
 		addr.sin_addr = so->so_faddr;
 	addr.sin_port = so->so_fport;
 	
@@ -471,7 +434,7 @@ tcp_fconnect(so)
  * connection is accepted, then do an accept() here.
  * But, a) we need to know who's trying to connect 
  * to the socket to be able to SYN the local-host, and
- * b) we are allready connected to the foregn host by
+ * b) we are already connected to the foreign host by
  * the time it gets to accept(), so... We simply accept
  * here and SYN the local-host.
  */ 
@@ -485,24 +448,16 @@ tcp_connect(inso)
 	struct tcpcb *tp;
 	int s, opt = 1;
 
-#ifdef DEBUG
-	debug_call(dfd,"tcp_connect(inso) called ...\n");
-	debug_call(dfd,"    inso = %d\n", (int)inso);
-	fflush_call(dfd);
-#endif
+	DEBUG_CALL("tcp_connect");
+	DEBUG_ARG("inso = %lx", (long)inso);
 	
 	/*
 	 * If it's an SS_ACCEPTONCE socket, no need to socreate()
 	 * another socket, just use the accept() socket.
 	 */
 	if (inso->so_state & SS_FACCEPTONCE) {
-		/* FACCEPTONCE allready have a tcpcb, so we just sbinit */
+		/* FACCEPTONCE already have a tcpcb */
 		so = inso;
-		if (sbinit(&so->so_rcv, tcp_rcvspace) == NULL ||
-		    sbinit(&so->so_snd, tcp_sndspace) == NULL) {
-			sofree(so);
-			return;
-		}
 	} else {
 		if ((so = socreate()) == NULL) {
 			/* If it failed, get rid of the pending connection */
@@ -510,7 +465,7 @@ tcp_connect(inso)
 			return;
 		}
 		if (tcp_attach(so) < 0) {
-			free(so); /* Not sofree */
+			free(so); /* NOT sofree */
 			return;
 		}
 		so->so_laddr = inso->so_laddr;
@@ -520,9 +475,6 @@ tcp_connect(inso)
 	(void) tcp_mss(sototcpcb(so), 0);
 
 	if ((s = accept(inso->s,(struct sockaddr *)&addr,&addrlen)) < 0) {
-#ifdef DEBUG
-		debug_misc(dfd," errno = %d\n", errno);
-#endif
 		tcp_close(sototcpcb(so)); /* This will sofree() as well */
 		return;
 	}
@@ -533,14 +485,14 @@ tcp_connect(inso)
 	so->so_fport = addr.sin_port;
 	so->so_faddr = addr.sin_addr;
 	/* Translate connections from localhost to the real hostname */
-	if (so->so_faddr.s_addr == 0 || so->so_faddr.s_addr == inet_addr("127.0.0.1"))
-	   so->so_faddr.s_addr = our_addr;
+	if (so->so_faddr.s_addr == 0 || so->so_faddr.s_addr == loopback_addr.s_addr)
+	   so->so_faddr = our_addr;
 	
 	/* Close the accept() socket, set right state */
 	if (inso->so_state & SS_FACCEPTONCE) {
-		close(so->s); /* If we only accept once, close the accet() socket */
+		close(so->s); /* If we only accept once, close the accept() socket */
 		so->so_state = SS_NOFDREF; /* Don't select it yet, even though we have an FD */
-					   /* if it's not FACCEPTONCE, it's allready NOFDREF */
+					   /* if it's not FACCEPTONCE, it's already NOFDREF */
 	}
 	so->s = s;
 	
@@ -568,53 +520,39 @@ tcp_connect(inso)
 
 /*
  * Attach a TCPCB to a socket.
- * Note: The caller is responsible of insque()ing the socket
- * to the relevent protocol CB head
  */
 int
 tcp_attach(so)
 	struct socket *so;
 {
-	if (sbinit(&so->so_rcv, tcp_rcvspace) == NULL)
-	 	return -1;
-	if (sbinit(&so->so_snd, tcp_sndspace) == NULL) {
-		sbfree(&so->so_rcv);
-		return -1;
-	}
-	if ((so->so_tcpcb = tcp_newtcpcb(so)) == NULL) {
-		sbfree(&so->so_rcv);
-		sbfree(&so->so_snd);
-		return -1;
-	}
+	if ((so->so_tcpcb = tcp_newtcpcb(so)) == NULL)
+	   return -1;
+	
 	insque(so, &tcb);
 
 	return 0;
 }
 
 /*
- * Set the socket's type of service feild
- * 
- * Any connection that needs emulation, has IPTOS_EMU set in it's
- * tos.. make sure we remove it before puting it in the TOS feild
- * of the IP header
+ * Set the socket's type of service field
  */
-struct tos_t {
-	u_int16_t lport;
-	u_int16_t fport;
-	u_int8_t tos;
-};
-
 struct tos_t tcptos[] = {
-	{0, 20, IPTOS_THROUGHPUT},		/* ftp data */
-	{21, 21, (IPTOS_LOWDELAY|IPTOS_EMU)},	/* ftp control */
-	{0, 23, IPTOS_LOWDELAY},			/* telnet */
-	{0, 80, IPTOS_THROUGHPUT},			/* WWW */
-	{0, 513, IPTOS_LOWDELAY},			/* rlogin */
-	{0, 6667, (IPTOS_THROUGHPUT|IPTOS_EMU)},	/* IRC */
-	{0, 6668, (IPTOS_THROUGHPUT|IPTOS_EMU)},	/* IRC undernet */
-	{0, 0, 0}
+	  {0, 20, IPTOS_THROUGHPUT, 0},	/* ftp data */
+	  {21, 21, IPTOS_LOWDELAY,  EMU_FTP},	/* ftp control */
+	  {0, 23, IPTOS_LOWDELAY, 0},	/* telnet */
+	  {0, 80, IPTOS_THROUGHPUT, 0},	/* WWW */
+	  {0, 513, IPTOS_LOWDELAY, EMU_RLOGIN|EMU_NOCONNECT},	/* rlogin */
+	  {0, 514, IPTOS_LOWDELAY, EMU_KSH},	/* shell */
+	  {0, 544, IPTOS_LOWDELAY, EMU_KSH},		/* kshell */
+	  {0, 543, IPTOS_LOWDELAY, 0},	/* klogin */
+	  {0, 6667, IPTOS_THROUGHPUT, EMU_IRC},	/* IRC */
+	  {0, 6668, IPTOS_THROUGHPUT, EMU_IRC},	/* IRC undernet */
+	  {0, 7070, IPTOS_LOWDELAY, EMU_REALAUDIO }, /* RealAudio control */
+	  {0, 0, 0, 0}
 };
 
+struct emu_t *tcpemu = 0;
+		
 /*
  * Return TOS according to the above table
  */
@@ -623,19 +561,26 @@ tcp_tos(so)
 	struct socket *so;
 {
 	int i = 0;
-
-#ifdef DEBUG
-	debug_call(dfd,"tcp_tos(so) called ...\n");
-	debug_call(dfd,"    so = %d\n", (int)so);
-	fflush_call(dfd);
-#endif
-
+	struct emu_t *emup;
+	
 	while(tcptos[i].tos) {
 		if ((tcptos[i].fport && (ntohs(so->so_fport) == tcptos[i].fport)) ||
-		    (tcptos[i].lport && (ntohs(so->so_lport) == tcptos[i].lport)))
+		    (tcptos[i].lport && (ntohs(so->so_lport) == tcptos[i].lport))) {
+			so->so_emu = tcptos[i].emu;
 			return tcptos[i].tos;
+		}
 		i++;
 	}
+	
+	/* Nope, lets see if there's a user-added one */
+	for (emup = tcpemu; emup; emup = emup->next) {
+		if ((emup->fport && (ntohs(so->so_fport) == emup->fport)) ||
+		    (emup->lport && (ntohs(so->so_lport) == emup->lport))) {
+			so->so_emu = emup->emu;
+			return emup->tos;
+		}
+	}
+	
 	return 0;
 }
 
@@ -662,9 +607,11 @@ int do_echo = -1;
  * 
  * Return 1 if the mbuf m is still valid and should be 
  * sbappend()ed
+ * 
+ * NOTE: if you return 0 you MUST m_free() the mbuf!
  */
 int
-tcp_emu(so,m)
+tcp_emu(so, m)
 	struct socket *so;
 	struct mbuf *m;
 {
@@ -672,102 +619,198 @@ tcp_emu(so,m)
 	char buff[256];
 	u_int32_t laddr;
 	u_int lport;
-	char *ptr;
-
-#ifdef DEBUG
-	debug_call(dfd,"tcp_emu(so,m) called ...\n");
-	debug_call(dfd,"    so = %d, m = %d\n", (int)so, (int)m);
-	fflush_call(dfd);
-#endif
+	char *bptr;
 	
-	/*
-	 * First, check if it's a CTL address so that we
-	 * can "execute" the "commands"
-	 */
-	if (ntohl(so->so_faddr.s_addr) == (ntohl(special_addr)|CTL_CMD)) {
-		int num;
-		struct sbuf *so_snd = &so->so_snd;
-		struct sbuf *so_rcv = &so->so_rcv;
+	DEBUG_CALL("tcp_emu");
+	DEBUG_ARG("so = %lx", (long)so);
+	DEBUG_ARG("m = %lx", (long)m);
+	
+	switch(so->so_emu) {
+		int x, i;
 		
+	 case EMU_RLOGIN:
 		/*
-		 * If there is binary data here, we save it in so->so_m
+		 * Rlogin emulation
+		 * First we accumulate all the initial option negotiation,
+		 * then fork_exec() rlogin according to the  options
 		 */
-		num = 0;
-		while (num++ < m->m_len) {
-			if (*mtod(m, char *) & 0x80) {
-				if (so->so_m == 0)
-				   so->so_m = m;
-				else
-				   m_free(m);
-				
+		{
+			int i, i2, n;
+			char *ptr;
+			char args[100];
+			char term[100];
+			struct sbuf *so_snd = &so->so_snd;
+			struct sbuf *so_rcv = &so->so_rcv;
+			
+			/* First check if they have a priveladged port, or too much data has arrived */
+			if (ntohs(so->so_lport) > 1023 || ntohs(so->so_lport) < 512 ||
+			    (m->m_len + so_rcv->sb_wptr) > (so_rcv->sb_data + so_rcv->sb_datalen)) {
+				memcpy(so_snd->sb_wptr, "Permission denied\n", 18);
+				so_snd->sb_wptr += 18;
+				so_snd->sb_cc += 18;
+				tcp_sockclosed(sototcpcb(so));
+				m_free(m);
 				return 0;
 			}
-		}
-		
-		/*
-		 * Append the line
-		 */
-		sbappendsb(so_rcv, m);
-		
-		/* To avoid going over the edge of the buffer, we reset it */
-		if (so_snd->sb_cc == 0)
-			so_snd->sb_wptr = so_snd->sb_rptr = so_snd->sb_data;
-		
-		/*
-		 * A bit of a hack:
-		 * If the first packet we get here is 1 byte long, then it
-		 * was done in telnet character mode, therefore we must echo
-		 * the characters as they come.  Otherwise, we echo nothing,
-		 * because in linemode, the line is allready echoed
-		 */
-		if (do_echo == -1) {
-			if (m->m_len == 1)
-				do_echo = 1;
-			else
-				do_echo = 0;
-		}
-		if (do_echo) {
-			sbappendsb(so_snd, m);
+			
+			/* Append the current data */
+			memcpy(so_rcv->sb_wptr, m->m_data, m->m_len);
+			so_rcv->sb_wptr += m->m_len;
+			so_rcv->sb_rptr += m->m_len;
 			m_free(m);
-			tcp_output(sototcpcb(so)); /* XXX */
-		} else
-			m_free(m);
-
-		num = 0;
-		while (num < so->so_rcv.sb_cc) {
-			if (*(so->so_rcv.sb_rptr + num) == '\n' ||
-			    *(so->so_rcv.sb_rptr + num) == '\r') {
-				int n;
-
-				*(so_rcv->sb_rptr + num) = 0;
-				n = do_config(so_rcv->sb_rptr, so, PRN_SPRINTF);
-				so_snd->sb_cc += n;
-				so_snd->sb_wptr += n;
-				/* Drop so_rcv data */
-				so_rcv->sb_cc = 0;
-				so_rcv->sb_wptr = so_rcv->sb_rptr = so_rcv->sb_data;
-				tcp_output(sototcpcb(so)); /* Send the reply */
-			}
-			num++;
-		}
-		
-		return 0;
-	}
-	
-	switch(ntohs(so->so_fport)) {
-		int x;
-
-	 case 21: /* ftp */
-ftp_emu:
-		*(m->m_data+m->m_len) = 0; /* NULL terminate for strstr */
-		if ((ptr = (char *)strstr(m->m_data, "ORT")) != NULL) {
-#ifdef DEBUG
-			debug_misc(dfd,"FTP - %s\n", ptr);
+			
+			/*
+			 * Check if we have all the initial options,
+			 * and build argument list to rlogin while we're here
+			 */
+			n = 0;
+			ptr = so_rcv->sb_data;
+			args[0] = 0;
+			term[0] = 0;
+			while (ptr < so_rcv->sb_wptr) {
+				if (*ptr++ == 0) {
+					n++;
+					if (n == 2) {
+						sprintf(args, "rlogin -l %s %s",
+							ptr, inet_ntoa(so->so_faddr));
+					} else if (n == 3) {
+						i2 = so_rcv->sb_wptr - ptr;
+						for (i = 0; i < i2; i++) {
+							if (ptr[i] == '/') {
+								ptr[i] = 0;
+#ifdef HAVE_SETENV
+								sprintf(term, "%s", ptr);
+#else
+								sprintf(term, "TERM=%s", ptr);
 #endif
+								ptr[i] = '/';
+								break;
+							}
+						}
+					}
+				}
+			}
+			
+			if (n != 4)
+			   return 0;
+			
+			/* We have it, set our term variable and fork_exec() */
+#ifdef HAVE_SETENV
+			setenv("TERM", term, 1);
+#else
+			putenv(term);
+#endif
+			fork_exec(so, args, 2);
+			term[0] = 0;
+			so->so_emu = 0;
+			
+			/* And finally, send the client a 0 character */
+			so_snd->sb_wptr[0] = 0;
+			so_snd->sb_wptr++;
+			so_snd->sb_cc++;
+			
+			return 0;
+		}
+		
+		
+	 case EMU_CTL:
+		{
+			int num;
+			struct sbuf *so_snd = &so->so_snd;
+			struct sbuf *so_rcv = &so->so_rcv;
+			
+			/*
+			 * If there is binary data here, we save it in so->so_m
+			 */
+			if (!so->so_m) {
+				for (num = 0; num < m->m_len; num++) {
+					if (*mtod(m, char *) & 0x80) {
+						so->so_m = m;
+						return 0;
+					}
+				}
+			}
+			
+			/*
+			 * Append the line
+			 */
+			sbappendsb(so_rcv, m);
+			
+			/* To avoid going over the edge of the buffer, we reset it */
+			if (so_snd->sb_cc == 0)
+			   so_snd->sb_wptr = so_snd->sb_rptr = so_snd->sb_data;
+			
+			/*
+			 * A bit of a hack:
+			 * If the first packet we get here is 1 byte long, then it
+			 * was done in telnet character mode, therefore we must echo
+			 * the characters as they come.  Otherwise, we echo nothing,
+			 * because in linemode, the line is already echoed
+			 * XXX two or more control connections won't work
+			 */
+			if (do_echo == -1) {
+				if (m->m_len == 1)
+				   do_echo = 1;
+				else
+				   do_echo = 0;
+			}
+			if (do_echo) {
+				sbappendsb(so_snd, m);
+				m_free(m);
+				tcp_output(sototcpcb(so)); /* XXX */
+			} else
+			   m_free(m);
+			
+			num = 0;
+			while (num < so->so_rcv.sb_cc) {
+				if (*(so->so_rcv.sb_rptr + num) == '\n' ||
+				    *(so->so_rcv.sb_rptr + num) == '\r') {
+					int n;
+					
+					*(so_rcv->sb_rptr + num) = 0;
+					if (ctl_password && !ctl_password_ok) {
+						/* Need a password */
+						if (sscanf(so_rcv->sb_rptr, "pass %256s", buff) == 1) {
+							if (strcmp(buff, ctl_password) == 0) {
+								ctl_password_ok = 1;
+								n = sprintf(so_snd->sb_wptr,
+									    "Password OK.\r\n");
+								goto do_prompt;
+							}
+						}
+						n = sprintf(so_snd->sb_wptr,
+					 "Error: Password required, log on with \"pass PASSWORD\"\r\n");
+						goto do_prompt;
+					}
+					cfg_quitting = 0;
+					n = do_config(so_rcv->sb_rptr, so, PRN_SPRINTF);
+					if (!cfg_quitting) {
+						/* Register the printed data */
+do_prompt:
+						so_snd->sb_cc += n;
+						so_snd->sb_wptr += n;
+						/* Add prompt */
+						n = sprintf(so_snd->sb_wptr, "Slirp> ");
+						so_snd->sb_cc += n;
+						so_snd->sb_wptr += n;
+					}
+					/* Drop so_rcv data */
+					so_rcv->sb_cc = 0;
+					so_rcv->sb_wptr = so_rcv->sb_rptr = so_rcv->sb_data;
+					tcp_output(sototcpcb(so)); /* Send the reply */
+				}
+				num++;
+			}
+			return 0;
+		}
+		
+	 case EMU_FTP: /* ftp */
+		*(m->m_data+m->m_len) = 0; /* NULL terminate for strstr */
+		if ((bptr = (char *)strstr(m->m_data, "ORT")) != NULL) {
 			/*
 			 * Need to emulate the PORT command
 			 */			
-			x = sscanf(ptr, "ORT %d,%d,%d,%d,%d,%d\r\n%256[^\177]", 
+			x = sscanf(bptr, "ORT %d,%d,%d,%d,%d,%d\r\n%256[^\177]", 
 				   &n1, &n2, &n3, &n4, &n5, &n6, buff);
 			if (x < 6)
 			   return 1;
@@ -790,18 +833,15 @@ ftp_emu:
 			n3 = ((laddr >> 8)  & 0xff);
 			n4 =  (laddr & 0xff);
 			
-			m->m_len = ptr - m->m_data; /* Adjust length */
-			m->m_len += sprintf(ptr,"ORT %d,%d,%d,%d,%d,%d\r\n%s", 
+			m->m_len = bptr - m->m_data; /* Adjust length */
+			m->m_len += sprintf(bptr,"ORT %d,%d,%d,%d,%d,%d\r\n%s", 
 					    n1, n2, n3, n4, n5, n6, x==7?buff:"");
-#ifdef DEBUG
-			debug_misc(dfd,"FTP2 - %s\n", ptr);
-#endif
 			return 1;
-		} else if ((ptr = (char *)strstr(m->m_data, "27 Entering")) != NULL) {
+		} else if ((bptr = (char *)strstr(m->m_data, "27 Entering")) != NULL) {
 			/*
 			 * Need to emulate the PASV response
 			 */
-			x = sscanf(ptr, "27 Entering Passive Mode (%d,%d,%d,%d,%d,%d)\r\n%256[^\177]",
+			x = sscanf(bptr, "27 Entering Passive Mode (%d,%d,%d,%d,%d,%d)\r\n%256[^\177]",
 				   &n1, &n2, &n3, &n4, &n5, &n6, buff);
 			if (x < 6)
 			   return 1;
@@ -824,8 +864,8 @@ ftp_emu:
 			n3 = ((laddr >> 8)  & 0xff);
 			n4 =  (laddr & 0xff);
 			
-			m->m_len = ptr - m->m_data; /* Adjust length */
-			m->m_len += sprintf(ptr,"27 Entering Passive Mode (%d,%d,%d,%d,%d,%d)\r\n%s",
+			m->m_len = bptr - m->m_data; /* Adjust length */
+			m->m_len += sprintf(bptr,"27 Entering Passive Mode (%d,%d,%d,%d,%d,%d)\r\n%s",
 					    n1, n2, n3, n4, n5, n6, x==7?buff:"");
 			
 			return 1;
@@ -833,47 +873,218 @@ ftp_emu:
 		
 		return 1;
 				   
-	 case 6667:
-	 case 6668:
+	 case EMU_KSH:
 		/*
-		 * Need to emulate DCC CHAT and DCC SEND
+		 * The kshell (Kerberos rsh) and shell services both pass
+		 * a local port port number to carry signals to the server
+		 * and stderr to the client.  It is passed at the beginning
+		 * of the connection as a NUL-terminated decimal ASCII string.
+		 */
+		so->so_emu = 0;
+		for (lport = 0, i = 0; i < m->m_len-1; ++i) {
+			if (m->m_data[i] < '0' || m->m_data[i] > '9')
+				return 1;       /* invalid number */
+			lport *= 10;
+			lport += m->m_data[i] - '0';
+		}
+		if (m->m_data[m->m_len-1] == '\0' && lport != 0 &&
+		    (so = solisten(0, so->so_laddr.s_addr, htons(lport), SS_FACCEPTONCE)) != NULL)
+			m->m_len = sprintf(m->m_data, "%d", ntohs(so->so_fport))+1;
+		return 1;
+		
+	 case EMU_IRC:
+		/*
+		 * Need to emulate DCC CHAT, DCC SEND and DCC MOVE
 		 */
 		*(m->m_data+m->m_len) = 0; /* NULL terminate the string for strstr */
-		if ((ptr = (char *)strstr(m->m_data, "DCC")) == NULL)
+		if ((bptr = (char *)strstr(m->m_data, "DCC")) == NULL)
 			 return 1;
 		
-		if (sscanf(ptr, "DCC CHAT chat %u %u", &laddr, &lport) == 2) {
+		/* The %256s is for the broken mIRC */
+		if (sscanf(bptr, "DCC CHAT %256s %u %u", buff, &laddr, &lport) == 3) {
 			if ((so = solisten(0, htonl(laddr), htons(lport), SS_FACCEPTONCE)) == NULL)
 				return 1;
 			
-			m->m_len = ptr - m->m_data; /* Adjust length */
-			m->m_len += sprintf(ptr, "DCC CHAT chat %lu %u%c\n", 
-			     ntohl(so->so_faddr.s_addr), ntohs(so->so_fport), 1);
+			m->m_len = bptr - m->m_data; /* Adjust length */
+			m->m_len += sprintf(bptr, "DCC CHAT chat %lu %u%c\n",
+			     (unsigned long)ntohl(so->so_faddr.s_addr),
+			     ntohs(so->so_fport), 1);
+		} else if (sscanf(bptr, "DCC SEND %256s %u %u %u", buff, &laddr, &lport, &n1) == 4) {
+			if ((so = solisten(0, htonl(laddr), htons(lport), SS_FACCEPTONCE)) == NULL)
+				return 1;
+			
+			m->m_len = bptr - m->m_data; /* Adjust length */
+			m->m_len += sprintf(bptr, "DCC SEND %s %lu %u %u%c\n", 
+			      buff, (unsigned long)ntohl(so->so_faddr.s_addr),
+			      ntohs(so->so_fport), n1, 1);
+		} else if (sscanf(bptr, "DCC MOVE %256s %u %u %u", buff, &laddr, &lport, &n1) == 4) {
+			if ((so = solisten(0, htonl(laddr), htons(lport), SS_FACCEPTONCE)) == NULL)
+				return 1;
+			
+			m->m_len = bptr - m->m_data; /* Adjust length */
+			m->m_len += sprintf(bptr, "DCC MOVE %s %lu %u %u%c\n",
+			      buff, (unsigned long)ntohl(so->so_faddr.s_addr),
+			      ntohs(so->so_fport), n1, 1);
+		}
+		return 1;
 
-			return 1;
-		}
-		if (sscanf(ptr, "DCC SEND %256s %u %u %u", buff, &laddr, &lport, &n1) == 4) {
-			if ((so = solisten(0, htonl(laddr), htons(lport), SS_FACCEPTONCE)) == NULL)
-				return 1;
+	case EMU_REALAUDIO:
+		/* 
+		 * RealAudio emulation - JP. We must try to parse the incoming
+		 * data and try to find the two characters that contain the
+		 * port number. Then we redirect an udp port and replace the
+		 * number with the real port we got.
+		 * Different RAplayer versions seem to use different message
+		 * formats. Here is an example of an 'old' packet:		 
+		 * 
+		 * 0000:50 
+		 * 0000:4E 41 00 04 63 00 12 50 4E 43 4C 49 45 4E 54 20 NA..c..PNCLIENT 
+		 * 0010:31 30 31 20 41 4C 50 48 41 6C 00 00 64 1B FD 52 101 ALPHAl..d.R
+		 * 0020:00 17 72 61 66 69 6C 65 73 2F 76 6F 61 2F 65 6E ..rafiles/voa/en
+		 * 0030:67 6C 69 73 68 5F 2E 72 61 79 42                glish_.rayB 
+		 *		
+		 * The port number 0x1BFD is found at displacement 0x1d of the second
+		 * packet (first packet contains only one byte 0x50).
+		 * 
+		 * Here is an example of the 'new' format:
+		 * 
+		 * 0000:50 4E 41 00 05 
+		 * 0000:00 01 00 02 1B D7 00 00 67 E6 6C DC 63 00 12 50 .......glc..P
+		 * 0010:4E 43 4C 49 45 4E 54 20 31 30 31 20 41 4C 50 48 NCLIENT 101 ALPH
+		 * 0020:41 6C 00 00 52 00 17 72 61 66 69 6C 65 73 2F 76 Al..R..rafiles/v
+		 * 0030:6F 61 2F 65 6E 67 6C 69 73 68 5F 2E 72 61 79 42 oa/english_.rayB
+		 * 
+		 * Now the port number 0x1BD7 is found at displacement 0x4 of the
+		 * second packet. This time we received five bytes first and then
+		 * the rest. You never know how many bytes will come.
+		 *
+		 * This is just a horrible switch statement. Variable ra tells
+		 * us where we go.
+		 */
+		bptr = m->m_data;
+		while (bptr < m->m_data + m->m_len){
+			u_short p;
+			static int ra = 0, textlen = 0;
+			char ra_tbl[6]; /* = {0x50, 0x4e, 0x41, 0, 0, 0x63}; */
 			
-			m->m_len = ptr - m->m_data; /* Adjust length */
-			m->m_len += sprintf(ptr, "DCC SEND %s %lu %u %u%c\n", 
-			      buff, ntohl(so->so_faddr.s_addr), ntohs(so->so_fport), n1, 1);
-			return 1;
+			ra_tbl[0] = 0x50;
+			ra_tbl[1] = 0x4e;
+			ra_tbl[2] = 0x41;
+			ra_tbl[3] = 0;
+			ra_tbl[4] = 0;
+			ra_tbl[5] = 0x63;
+			
+			switch (ra) {
+			 case 0:
+			 case 1:
+			 case 2:
+			 case 3:
+			 case 5:
+				if (*bptr++ != ra_tbl[ra]) {
+					ra = 0;
+					return 1;
+				} 
+				break;
+
+			/*
+			 * Now the 5th char tells us which format to use.
+			 * It was 4 for the BETAs, but is 5 for the release
+			 * version 1.0.0. I hope they don't change the for-
+			 * mat in every new version.
+			 * If this char is 5, just add 100 to ra and con-
+			 * tinue from case 105.
+			 */
+
+			 case 4:
+				if (*bptr == 4)
+					bptr++;
+				else if (*bptr == 5){
+					bptr++;
+					ra += 100;
+				} else
+					ra = 0;	
+				break;
+			/*
+			 * OK, we're using the old format. Here we should
+			 * have the length of RAplayers name.
+			 */
+			 case 6: 
+				textlen  = *(u_char *)bptr++ << 8;
+				textlen += *(u_char *)bptr++;
+				break;
+
+			/* 
+			 * Just skip the name
+			 */ 
+			 case 7:
+				bptr += textlen;
+				break;
+			 case 8:				/* expect 0x6c */
+				if (*bptr++ != 0x6c){
+					ra = 0;
+					return 1;
+				}
+				break;
+			 case 9:
+				/* Not sure what this field means, seems
+				 * to be two zeros usually. Just get
+				 * past it.
+				 */
+				 bptr += 2;
+				 break;
+			 case 10:			/* expect 0x64 */
+				if (*bptr++ != 0x64){
+					ra = 0;
+					return 1;
+				}
+				break;
+				
+			 case 11:
+			 case 106:
+				/* This is the field containing the port
+				 * number that RA-player is listening to.
+				 */
+				 lport = (((u_char*)bptr)[0] << 8) 
+				 	 + ((u_char *)bptr)[1];
+				 if (lport < 6970) 	
+				 	lport += 256; 	/* don't know why */
+				 /* try to get udp port between 6970 - 7170 */
+				 for (p = 6970; p < 7071; p++){
+				 	if (udp_listen( htons(p),
+				 			so->so_laddr.s_addr,
+				 			htons(lport),
+				 			SS_FACCEPTONCE)){
+				 				break;
+				 	}
+				 }
+				 if (p == 7071)
+				 	p = 0;
+				 *(u_char *)bptr++ = (p >> 8) & 0xff;
+				 *(u_char *)bptr++ = p & 0xff;
+				ra = 0; 
+				return 1;   /* port redirected, we're done */
+				break;	
+
+			/* 
+			 * This is the new message format. Expecting 
+			 * 0,1,0,2, but just skipping over it to get
+			 * to the port number field. It is handled in
+			 * case 106 above.
+			 */				
+			 case 105:
+				bptr += 4;				 
+				break;
+				
+			}
+			ra++;
 		}
-		
+		return 1;			    	 
+
+	 default:
+		/* Ooops, not emulated, won't call tcp_emu again */
+		so->so_emu = 0;
 		return 1;
 	}
-	
-	/* Not fport, must be lport */
-	switch(ntohs(so->so_lport)) {
-	 case 21:
-		goto ftp_emu;
-	 default:
-		break;
-	}
-	
-	return 1;
 }
 
 /*
@@ -889,18 +1100,16 @@ tcp_ctl(so)
 	int command;
 	struct ex_list *ex_ptr;
 	int do_pty;
+	struct socket *tmpso;
 	
-#ifdef DEBUG
-	debug_call(dfd,"tcp_ctl(so) called ...\n");
-	debug_call(dfd,"    so = %d\n", (int)so);
-	fflush_call(dfd);
-#endif
+	DEBUG_CALL("tcp_ctl");
+	DEBUG_ARG("so = %lx", (long)so);
 	
 	/*
 	 * Check if they're authorised
 	 */
-	if (ctl_addr && (ctl_addr == -1 || (so->so_laddr.s_addr != ctl_addr))) {
-		sb->sb_cc = sprintf(sb->sb_wptr," Error: Permission denied.\n");
+	if (ctl_addr.s_addr && (ctl_addr.s_addr == -1 || (so->so_laddr.s_addr != ctl_addr.s_addr))) {
+		sb->sb_cc = sprintf(sb->sb_wptr," Error: Permission denied.\r\n");
 		sb->sb_wptr += sb->sb_cc;
 		return 0;
 	}
@@ -922,9 +1131,7 @@ tcp_ctl(so)
 		}
 		
 		/*
-		 * Nothing bound.. should never happen, except in a bizarre
-		 * race condition (when user removes an "exec" while
-		 * this session was connecting)
+		 * Nothing bound..
 		 */
 		tcp_fconnect(so);
 		
@@ -937,16 +1144,19 @@ do_exec:
 		return fork_exec(so, ex_ptr->ex_exec, do_pty);
 		
 	 case CTL_CMD:
-		/*
-		 * This is an aweful hack
-		 * We set the session IPTOS_EMU so that tcp_emu gets called,
-		 * so that we can "execute" the commands given, in tcp_emu.
-		 * Yuck! :)
-		 */
-		so->so_iptos |= IPTOS_EMU;
-		sb->sb_cc = sprintf(sb->sb_wptr, " SLiRP command-line ready (type \"help\" for help).\n");
+		for (tmpso = tcb.so_next; tmpso != &tcb; tmpso = tmpso->so_next) {
+			if (tmpso->so_emu == EMU_CTL &&
+			    !(tmpso->so_tcpcb?(tmpso->so_tcpcb->t_state & TCPS_TIME_WAIT):0)) {
+				/* Ooops, control connection already active */
+				sb->sb_cc = sprintf(sb->sb_wptr," Sorry, already connected.\r\n");
+				sb->sb_wptr += sb->sb_cc;
+				return 0;
+			}
+		}
+		so->so_emu = EMU_CTL;
+		ctl_password_ok = 0;
+		sb->sb_cc = sprintf(sb->sb_wptr, " SLiRP command-line ready (type \"help\" for help).\r\nSlirp> ");
 		sb->sb_wptr += sb->sb_cc;
 		return 2;
-
 	}
 }
